@@ -64,6 +64,16 @@ fn bench(c: &mut Criterion) {
   vips.concurrency_set(1);
   let vip = libvips::VipsImage::new_from_file("tests/images/milk.png").unwrap().image_decode().unwrap();
 
+  c.benchmark_group("quant::quantize")
+    .bench_function("imagescript::quant::quantize", |b| b.iter(|| imagescript::quant::quantize(&mut fb, &imagescript::quant::options { ..Default::default() })))
+
+    .bench_function("image-rs::quant::quantize", |b| b.iter(|| {
+      let q = color_quant::NeuQuant::new(1, 256, &fb);
+      let v: Vec<_> = (&fb).chunks_exact(4).map(|pix| q.index_of(pix) as u8).collect();
+
+      return (q.color_map_rgb(), v);
+    }));
+
   c.benchmark_group("ops::fill")
     .bench_function("image-rs::color", |b| b.iter(|| imagers_fill(&mut image)))
     .bench_function("imagescript::color", |b| b.iter(|| imagescript_fill(&mut fb)));
