@@ -5,6 +5,7 @@
 #![feature(unchecked_math)]
 #![warn(clippy::complexity)]
 #![feature(core_intrinsics)]
+#![feature(const_ptr_offset)]
 #![warn(clippy::correctness)]
 #![allow(non_camel_case_types)]
 #![feature(downcast_unchecked)]
@@ -33,7 +34,7 @@ unsafe impl Send for framebuffer {}
 unsafe impl Sync for framebuffer {}
 
 pub trait framebuffer_from<T> {
-  fn from(width: usize, height: usize, t: T) -> fb;
+  fn from(width: usize, height: usize, container: T) -> fb;
 }
 
 impl Drop for framebuffer {
@@ -90,8 +91,8 @@ impl framebuffer {
 }
 
 pub mod colors {
-  #[repr(C)] #[derive(Copy, Clone)] pub struct rgb { pub r: u8, pub g: u8, pub b: u8 }
-  #[repr(C)] #[derive(Copy, Clone)] pub struct rgba { pub r: u8, pub g: u8, pub b: u8, pub a: u8 }
+  #[repr(C)] #[derive(Copy, Clone, Debug)] pub struct rgb { pub r: u8, pub g: u8, pub b: u8 }
+  #[repr(C)] #[derive(Copy, Clone, Debug)] pub struct rgba { pub r: u8, pub g: u8, pub b: u8, pub a: u8 }
 
   impl From<rgba> for rgb { #[inline] fn from(c: rgba) -> rgb { return rgb { r: c.r, g: c.g, b: c.b }; } }
   impl From<rgb> for rgba { #[inline] fn from(c: rgb) -> rgba { return rgba { r: c.r, g: c.g, b: c.b, a: 255 }; } }
@@ -845,4 +846,422 @@ pub mod ops {
       }
     }
   }
+}
+
+pub mod quant {
+  use super::*;
+
+  struct bin { r: u32, g: u32, b: u32, a: u32, count: u32 }
+
+  struct tbin {
+    n: u32, p: u32,
+    r: f32, g: f32, b: f32, a: f32,
+    la: u32, lb: u32, error: f32, count: f32, nearest: u32,
+  }
+
+  pub struct result {
+    pub index: Vec<u8>,
+    pub palette: Vec<colors::rgba>,
+    pub transparency: Option<usize>,
+  }
+
+  pub struct options {
+    pub fast: bool,
+    pub dither: bool,
+    pub colors: usize,
+    pub transparency: bool,
+    pub threshold: Option<u8>,
+    pub force_transparency: bool,
+  }
+
+  impl Default for options {
+    fn default() -> Self {
+      return Self {
+        fast: false,
+        colors: 256,
+        dither: true,
+        threshold: None,
+        transparency: true,
+        force_transparency: false,
+      };
+    }
+  }
+
+  unsafe fn index(fb: &fb, _dither: bool, _colors: usize, palette: &[colors::rgba], s: &(f32, f32, f32, f32)) -> Vec<u8> {
+    let mut index: Vec<u8> = Vec::with_capacity(fb.width * fb.height);
+
+    let u8 = index.as_mut_ptr();
+    let u32 = fb.as_ptr::<u32>();
+    index.set_len(fb.width * fb.height);
+    let mut nearest: fnv::FnvHashMap<u32, u8> = fnv::FnvHashMap::default();
+    let mut closest: fnv::FnvHashMap<u32, (u8, u8, f32, f32)> = fnv::FnvHashMap::default();
+
+    for o in 0..(fb.len() / 4) {
+      *u8.add(o) = self::closest(*u32.add(o), o, &palette, &s, &mut nearest, &mut closest);
+    }
+
+    return index;
+  }
+
+  unsafe fn link_neareast(o: usize, bins: *mut tbin, s: &(f32, f32, f32, f32)) {
+    let mut nearest = 0;
+    let mut e = f32::INFINITY;
+    let bin = &mut *bins.add(o);
+
+    let n1 = bin.count;
+    let mut offset = bin.n;
+    let w = (bin.r, bin.g, bin.b, bin.a);
+
+    while 0 != offset {
+      let b = &*bins.add(offset as usize);
+
+      let mut err =
+        s.0 * (b.r - w.0).powi(2)
+        + s.1 * (b.g - w.1).powi(2)
+        + s.2 * (b.b - w.2).powi(2)
+        + s.3 * (b.a - w.3).powi(2);
+
+      let n2 = b.count;
+      err *= (n1 * n2) / (n1 + n2);
+      if e > err { e = err; nearest = offset; }
+      offset = (*bins.offset(offset as isize)).n;
+    }
+
+    bin.error = e;
+    bin.nearest = nearest;
+  }
+
+  unsafe fn nearest(c: u32, palette: &[colors::rgba], s: &(f32, f32, f32, f32), cache: &mut fnv::FnvHashMap<u32, u8>) -> u8 {
+    let old = cache.get(&c);
+    if old.is_some() { return *old.unwrap_unchecked(); }
+
+    let a = (c >> 24) as f32;
+    let r = (c & 0xff) as f32;
+    let g = ((c >> 8) & 0xff) as f32;
+    let b = ((c >> 16) & 0xff) as f32;
+
+    let mut offset = 0;
+    let p = palette.as_ptr();
+    let mut m = f32::INFINITY;
+
+    for o in 0..palette.len() {
+      let cc = *p.add(o);
+      let rr = cc.r as f32;
+      let gg = cc.g as f32;
+      let bb = cc.b as f32;
+      let aa = cc.a as f32;
+      let mut d = s.3 * (aa - a).powi(2);
+
+      if d > m { continue; }
+      d += s.0 * (rr - r).powi(2); if d > m { continue; }
+      d += s.1 * (gg - g).powi(2); if d > m { continue; }
+      d += s.2 * (bb - b).powi(2); if d > m { continue; }
+
+      m = d; offset = o as u8;
+    }
+
+    cache.insert(c, offset); return offset;
+  }
+
+  unsafe fn closest(c: u32, o: usize, palette: &[colors::rgba], s: &(f32, f32, f32, f32), nearest: &mut fnv::FnvHashMap<u32, u8>, cache: &mut fnv::FnvHashMap<u32, (u8, u8, f32, f32)>) -> u8 {
+    let old = cache.get(&c);
+
+    let m = if old.is_some() { *old.unwrap_unchecked() } else {
+      let a = (c >> 24) as f32;
+      let r = (c & 0xff) as f32;
+      let pp = palette.as_ptr();
+      let g = ((c >> 8) & 0xff) as f32;
+      let b = ((c >> 16) & 0xff) as f32;
+      let mut m = (0u8, 0u8, f32::INFINITY, f32::INFINITY);
+
+      for o in 0..palette.len() {
+        let cc = *pp.add(o);
+        let rr = cc.r as f32;
+        let gg = cc.g as f32;
+        let bb = cc.b as f32;
+        let aa = cc.a as f32;
+        let err = s.0 * (rr - r).powi(2) + s.1 * (gg - g).powi(2) + s.2 * (bb - b).powi(2) + s.3 * (aa - a).powi(2);
+        if m.2 > err { m.1 = m.0; m.3 = m.2; m.2 = err; m.0 = o as u8; } else if m.3 > err { m.3 = err; m.1 = o as u8; }
+      }
+
+      if m.3 == f32::INFINITY { m.1 = m.0; } cache.insert(c, m); m
+    };
+
+    let l = palette.len();
+    let mut oo = (1 + o) % 2;
+    if m.3 * 0.67 < (m.3 - m.2) { oo = 0 } else if m.0 > m.1 { oo = o % 2; }
+
+    let err = match oo { 0 => m.2, 1 => m.3, _ => unreachable() };
+    if err >= l as f32 { return self::nearest(c, palette, s, nearest); }
+
+    return match oo { 0 => m.0, 1 => m.1, _ => unreachable() };
+  }
+
+  pub fn quantize(fb: &fb, options: &options) -> result {
+    let width = fb.width;
+    let height = fb.height;
+    let mut fb = fb.clone();
+    let u32 = fb.as_mut_ptr::<u32>();
+    let colors = options.colors.clamp(2, 256);
+    let mut s = (0.2126, 0.7152, 0.0722, 0.3333);
+
+    if 32 >= colors { s = (1.0, 1.0, 1.0, 1.0); }
+    else if 512 > width || 512 > height { s = (0.299, 0.587, 0.114, s.3); }
+
+    unsafe {
+      if !options.transparency {
+        s.3 = 0.0;
+        let threshold = options.threshold.unwrap_or(0x42) as u32;
+
+        for o in 0..(fb.len() / 4) {
+          let a = *u32.add(o) >> 24;
+          *u32.add(0) = (if a < threshold { 0 } else { *u32.add(o) & 0xffffff }) | (255 << 24);
+        }
+      } else {
+        if options.threshold.is_none() {
+          for o in 0..(fb.len() / 4) {
+            let a = *u32.add(o) >> 24;
+            if a == 0 { *u32.add(0) = 0xffffff; }
+          }
+        } else {
+          let threshold = options.threshold.unwrap_unchecked() as u32;
+
+          for o in 0..(fb.len() / 4) {
+            let a = *u32.add(o) >> 24;
+            *u32.add(0) = if a < threshold { 0xffffff } else { *u32.add(o) & 0xffffff | (255 << 24) };
+          }
+        }
+      }
+    }
+
+    let mut palette: Vec<colors::rgba> = vec![0.into(); colors];
+
+    if 2 < colors {
+      let (mut s, p) = unsafe { self::palette(&fb, colors, &options, &mut s) };
+
+      unsafe {
+        let ss = s;
+        let mut t = -1;
+        for o in 0..s { if 0 == (*p.add(o) >> 24) { t = o as isize; break; } }
+        if t == -1 && options.transparency && options.force_transparency { *p.add(if s == colors { s - 1 } else { s += 1; s - 1 }) = 0; }
+
+        palette.as_mut_ptr().copy_from_nonoverlapping(p as _, s); palette.set_len(s);
+
+        free(p as *mut u8, ss);
+      }
+    } else {
+      unsafe {
+        *palette.get_unchecked_mut(0) = colors::rgba { r: 0, g: 0, b: 0, a: 255 };
+
+        let mut t = options.transparency && options.force_transparency;
+        if !t && options.transparency { for o in 0..(fb.len() / 4) { if 0 == (*u32.add(o) >> 24) { t = true; break; } } }
+
+        *palette.get_unchecked_mut(1) =
+          if t { colors::rgba { r: 255, g: 255, b: 255, a: 0 } }
+          else { colors::rgba { r: 255, g: 255, b: 255, a: 255 } };
+      }
+    }
+
+    let mut t = None;
+    let index = unsafe { index(&fb, options.dither, colors, &palette, &s) };
+    for o in 0..palette.len() { if 0 == palette[o].a { t = Some(o); break; } }
+
+    return result {
+      index,
+      palette,
+      transparency: t,
+    };
+  }
+
+  unsafe fn palette(fb: &fb, colors: usize, options: &options, s: &mut (f32, f32, f32, f32)) -> (usize, *mut u32) {
+    let u32 = fb.as_ptr::<u32>();
+    const BINS: usize = 2usize.pow(16);
+
+    let (size, bins): (usize, *mut tbin) = {
+      // copy v8 array design?
+      let bins = calloc(BINS * size_of!(bin)) as *mut bin;
+
+      let index = match options.fast {
+        true => |r: u32, g: u32, b: u32, a: u32| ((r & 0xF0) << 4) | (g & 0xF0) | (b >> 4) | ((a & 0xF0) << 8),
+
+        false => match 64 <= colors && !options.transparency {
+          true => |r: u32, g: u32, b: u32, _: u32| ((r & 0xf8) << 8) | ((g & 0xfc) << 3) | (b >> 3),
+          false => |r: u32, g: u32, b: u32, a: u32| ((r & 0xf8) << 7) | ((g & 0xf8) << 2) | (b >> 3) | ((a & 0x80) << 8),
+        }
+      };
+
+      for o in 0..(fb.len() / 4) {
+        let c = *u32.add(o);
+
+        let a = c >> 24;
+        let r = c & 0xff;
+        let g = (c >> 8) & 0xff;
+        let b = (c >> 16) & 0xff;
+
+        let mut bin = &mut *bins.add(index(r, g, b, a) as usize);
+
+        bin.r += r;
+        bin.g += g;
+        bin.b += b;
+        bin.a += a;
+        bin.count += 1;
+      }
+
+      let mut size = 0;
+
+      for o in 0..BINS {
+        let bin = bins.add(o);
+        if 0 == (*bin).count { continue; }
+        bins.add(size).copy_from(bin, 1); size += 1;
+      }
+
+      let tbin = calloc(size * size_of!(tbin)) as *mut tbin;
+
+      for o in 0..size {
+        use std::intrinsics::{fdiv_fast as fd};
+
+        let bin = &*bins.add(o);
+        let tbin = &mut *tbin.add(o);
+
+        let count = bin.count as f32;
+
+        tbin.count = count;
+        tbin.r = fd(bin.r as f32, count);
+        tbin.g = fd(bin.g as f32, count);
+        tbin.b = fd(bin.b as f32, count);
+        tbin.a = fd(bin.a as f32, count);
+      }
+
+      free(bins as *mut u8, BINS * size_of!(bin));
+
+      (size, tbin)
+    };
+
+    let q = {
+      let weight = colors as f32 / size as f32;
+      let mut e = if 16 <= colors { 1 } else { -1 };
+      if 0.003 < weight && 0.005 > weight { e = 0; };
+
+      if s.1 < 1.0 && 0.025 > weight {
+        s.1 -= 3.0 * (0.025 + weight);
+        s.2 += 3.0 * (0.025 + weight);
+      }
+
+      match e {
+        0 => |count: f32| count,
+        -1 => |count: f32| count.cbrt() as u32 as f32,
+  
+        1 => match 64 > colors {
+          true => |count: f32| count.sqrt(),
+          false => |count: f32| count.sqrt() as u32 as f32,
+        },
+
+        _ => unreachable(),
+      }
+    };
+
+    for o in 0..(size - 1) {
+      let bin = &mut *bins.add(o);
+
+      bin.n = (o + 1) as u32;
+      bin.count = q(bin.count);
+      (*bins.add(o + 1)).p = o as u32;
+    }
+
+    (*bins.add(size - 1)).count = q((*bins.add(size - 1)).count);
+
+    let heap = calloc(4 * (1 + BINS)) as *mut u32;
+
+    for o in 0..size {
+      link_neareast(o, bins, &s);
+      let err = (*bins.add(o)).error;
+
+      *heap += 1;
+      let mut l = *heap;
+
+      while 1 < l {
+        let h = *heap.add((l >> 1) as usize);
+        if err >= (*bins.add(h as usize)).error { break; }
+
+        *heap.add(l as usize) = h; l >>= 1;
+      }
+
+      *heap.add(l as usize) = o as u32;
+    }
+
+    let ext = size as isize - colors as isize;
+
+    for o in 0..ext {
+      let mut b;
+
+      loop {
+        b = *heap.add(1);
+        let tb = &mut *bins.add(b as usize);
+        if (tb.la >= tb.lb) && (tb.la >= (*bins.add(tb.nearest as usize)).lb) { break; }
+
+        if tb.lb == 0xffff {
+          *heap -= 1;
+          let h = 1 + *heap;
+          b = *heap.add(h as usize);
+          *heap.offset(1) = *heap.add(h as usize);
+        } else { link_neareast(b as usize, bins, &s); tb.la = o as u32; }
+
+        let mut l = 1;
+        let err = (*bins.add(b as usize)).error;
+
+        while l + l <= *heap {
+          let mut l2 = l + l;
+          if (l2 < *heap) && ((*bins.add(*heap.add(l2 as usize) as usize)).error > (*bins.add(*heap.add(1 + l2 as usize) as usize)).error) { l2 += 1 }
+
+          let h = *heap.add(l2 as usize);
+          if err <= (*bins.add(h as usize)).error { break; }
+
+          *heap.add(l as usize) = h; l = l2;
+        }
+
+        *heap.add(l as usize) = b;
+      }
+
+      let tb = &mut *bins.add(b as usize);
+      let nb = &mut *bins.add(tb.nearest as usize);
+
+      let n1 = tb.count;
+      let n2 = nb.count;
+      let d = 1.0 / (n1 + n2);
+      tb.r = d * (n1 * tb.r + n2 * nb.r).round();
+      tb.g = d * (n1 * tb.g + n2 * nb.g).round();
+      tb.b = d * (n1 * tb.b + n2 * nb.b).round();
+      tb.a = d * (n1 * tb.a + n2 * nb.a).round();
+
+      nb.lb = 0xffff;
+      tb.count += nb.count;
+      tb.lb = (o + 1) as u32;
+
+      (*bins.add(nb.p as usize)).n = nb.n;
+      (*bins.add(nb.n as usize)).p = nb.p;
+    }
+
+    free(heap as *mut u8, 4 * (1 + BINS));
+    let s = if 0 > ext { size } else { colors };
+    let palette = alloc(s * size_of!(colors::rgba)) as *mut colors::rgba;
+
+    let mut w = 0;
+    let mut offset = 0;
+
+    loop {
+      let r = (*bins.add(offset)).r.clamp(0.0, 255.0) as u8;
+      let g = (*bins.add(offset)).g.clamp(0.0, 255.0) as u8;
+      let b = (*bins.add(offset)).b.clamp(0.0, 255.0) as u8;
+      let a = (*bins.add(offset)).a.clamp(0.0, 255.0) as u8;
+
+      *palette.add(w) = colors::rgba { r, g, b, a };
+
+      w += 1;
+      offset = (*bins.add(offset)).n as usize;
+
+      if 0 == offset { break; }
+    }
+
+    return (s, palette as *mut u32);
+  }
 }
\ No newline at end of file
